"""Fixtures for supporting a distributed test run."""
import os
import shutil

import psutil
import pytest


@pytest.fixture(scope='session')
def root_dir():
    """The parent directory of the test base artifact directory."""
    return '/tmp'


@pytest.yield_fixture(scope='session')
def base_dir(request, session_id, root_dir, services_log):
    """The directory where test run artifacts should be stored.

    It is responsibility of fixtures and tests that depend on it to clean up
    after themselves.

    """
    path = os.path.join(root_dir, 'sr-{0}'.format(session_id))
    services_log.debug('creating base dir: {0}'.format(path))
    if not os.path.exists(path):
        os.mkdir(path)

    yield path

    services_log.debug('finalizing base dir: {0}'.format(path))
    shutil.rmtree(path, ignore_errors=True)


@pytest.fixture(scope='session')
def temp_dir(request, base_dir, services_log):
    """The temporary dir."""
    path = os.path.join(base_dir, 'tmp')

    services_log.debug('creating temp dir: {0}'.format(path))
    if not os.path.exists(path):
        os.mkdir(path)

    return path


@pytest.fixture(scope='session')
def memory_root_dir(root_dir):
    """The parent directory of the test artifact directory in memory."""
    # check for a free space for at least 8 parallel processes
    if os.path.exists('/dev/shm') and psutil.disk_usage('/dev/shm').free > 1024 * 1024 * 64 * 10:
        return '/dev/shm'
    return root_dir


@pytest.yield_fixture(scope='session')
def memory_base_dir(request, session_id, memory_root_dir, services_log):
    """The directory where memory test run artifacts should be stored.

    It is responsibility of fixtures and tests that depend on it to clean up
    after themselves.

    """
    path = os.path.join(memory_root_dir, 'sr-{0}'.format(session_id))

    services_log.debug('creating memory base dir: {0}'.format(path))
    if not os.path.exists(path):
        os.mkdir(path)

    yield path

    services_log.debug('finalizing memory base dir: {0}'.format(path))
    shutil.rmtree(path, ignore_errors=True)


@pytest.fixture(scope='session')
def memory_temp_dir(request, memory_base_dir, services_log):
    """The memory temporary dir."""
    path = os.path.join(memory_base_dir, 'tmp')

    services_log.debug('creating memory temp dir: {0}'.format(path))
    if not os.path.exists(path):
        os.mkdir(path)

    return path


@pytest.fixture(scope='session')
def lock_dir(memory_root_dir, services_log):
    """The lock dir."""
    path = os.path.join(memory_root_dir, 'service-locks')
    services_log.debug('ensuring lock dir: {0}'.format(path))
    if not os.path.exists(path):
        try:
            os.mkdir(path, 0o777)
        except OSError:
            # concurrent already created this path
            pass
    try:
        os.chmod(path, 0o777)  # permissions on previous line can be ignored
    except OSError:
        # could happen, but we'll continue as it's just a folder permissions
        pass

    return path


@pytest.fixture(scope='session')
def run_dir(memory_temp_dir, services_log):
    """The run dir (like local /var/run)."""
    path = os.path.join(memory_temp_dir, 'run')
    services_log.debug('creating run dir: {0}'.format(path))
    os.mkdir(path)

    return path



"""Fixtures for memcache."""
import os
import pytest


@pytest.fixture(scope='session')
def memcached_socket(run_dir, run_services):
    """The memcached socket location."""
    if run_services:
        return os.path.join(run_dir, 'memcached.sock')


@pytest.fixture(scope='session')
def memcached(request, run_services, memcached_socket, watcher_getter):
    """The memcached instance which is ready to be used by the tests."""
    if run_services:
        return watcher_getter(
            name='memcached',
            arguments=['-s', memcached_socket],
            checker=lambda: os.path.exists(memcached_socket),
            request=request,
        )


@pytest.fixture(scope='session')
def memcached_connection(run_services, memcached_socket):
    """The connection string to the local memcached instance."""
    if run_services:
        return 'unix:{0}'.format(memcached_socket)


@pytest.fixture
def do_memcached_clean(run_services):
    """Determine whether memcached should be clean on the start of every test."""
    return run_services


@pytest.fixture(scope='session')
def memcached_client(memcached_socket, memcached):
    """Create client for memcached."""
    mc = pytest.importorskip('pylibmc')
    return mc.Client([memcached_socket])


@pytest.fixture
def memcached_clean(request, memcached_client, do_memcached_clean):
    """Clean memcached instance."""
    if do_memcached_clean:
        memcached_client.flush_all()



"""Fixtures for supporting a distributed test run."""
import contextlib
import json
import os
from random import random
import socket
import time

import pytest
import zc.lockfile

marker = object()


def try_remove(filename):
    try:
        os.unlink(filename)
    except OSError:
        pass


@contextlib.contextmanager
def file_lock(filename, remove=True, timeout=20):
    """A lock that is shared across processes.

    :param filename: the name of the file that will be locked.
    :param remove: whether or not to remove the file on context close
    :param timeout: Amount of time to retry the file lock if a :class:`zc.lockfile.LockError` is hit
    """
    total_seconds_slept = 0
    while True:
        try:
            with contextlib.closing(zc.lockfile.SimpleLockFile(filename)) as lockfile:
                yield lockfile._fp
                break
        except zc.lockfile.LockError as err:
            if total_seconds_slept >= timeout:
                raise err
        seconds_to_sleep = random() * 0.1 + 0.05
        total_seconds_slept += seconds_to_sleep
        time.sleep(seconds_to_sleep)

    remove and try_remove(filename)


def unlock_resource(name, resource, lock_dir, services_log):
    """Unlock previously locked resource.

    :param name: name to be used to separate various resources, eg. port, display
    :param resource: resource value which was previously locked
    :param lock_dir: directory for lockfiles to use.
    """
    with locked_resources(name, lock_dir) as bound_resources:
        try:
            bound_resources.remove(resource)
        except ValueError:
            pass
        services_log.debug('resource freed {0}: {1}'.format(name, resource))
        services_log.debug('bound resources {0}: {1}'.format(name, bound_resources))


@contextlib.contextmanager
def locked_resources(name, lock_dir):
    """Contextmanager providing an access to locked shared resource list.

    :param name: name to be used to separate various resources, eg. port, display
    :param lock_dir: directory for lockfiles to use.
    """
    with file_lock(os.path.join(lock_dir, name), remove=False) as fd:
        bound_resources = fd.read().strip()
        if bound_resources:
            try:
                bound_resources = json.loads(bound_resources)
            except ValueError:
                bound_resources = None
        if not isinstance(bound_resources, list):
            bound_resources = []
        yield bound_resources

        fd.seek(0)
        fd.truncate()
        fd.write(json.dumps(bound_resources))
        fd.flush()


def unlock_port(port, lock_dir, services_log):
    """Unlock previously locked port."""
    return unlock_resource('port', port, lock_dir, services_log)


def unlock_display(display, lock_dir, services_log):
    """Unlock previously locked display."""
    return unlock_resource('display', display, lock_dir, services_log)


@pytest.fixture(scope='session')
def lock_resource_timeout():
    """Max number of seconds to obtain the lock."""
    return 20


def lock_resource(name, resource_getter, lock_dir, services_log, lock_resource_timeout):
    """Issue a lock for given resource."""
    total_seconds_slept = 0
    while True:
        try:
            with locked_resources(name, lock_dir) as bound_resources:
                services_log.debug('bound_resources {0}: {1}'.format(name, bound_resources))
                resource = resource_getter(bound_resources)
                while resource in bound_resources:
                    # resource is already taken by someone, retry
                    services_log.debug('bound resources {0}: {1}'.format(name, bound_resources))
                    resource = resource_getter(bound_resources)
                services_log.debug('free resource choosen {0}: {1}'.format(name, resource))
                bound_resources.append(resource)
                services_log.debug('bound resources {0}: {1}'.format(name, bound_resources))
                return resource
        except zc.lockfile.LockError as err:
            if total_seconds_slept >= lock_resource_timeout:
                raise err
            services_log.debug('lock resource failed: {0}'.format(err))

        seconds_to_sleep = random() * 0.1 + 0.05
        total_seconds_slept += seconds_to_sleep
        time.sleep(seconds_to_sleep)


def get_free_port(lock_dir, services_log, lock_resource_timeout):
    """Get free port to listen on."""
    def get_port(bound_resources):
        if bound_resources:
            port = max(bound_resources) + 1
        else:
            port = 30000

        while True:
            s = socket.socket()
            try:
                s.bind(('127.0.0.1', port))
                s.close()
                return port
            except socket.error:
                pass
            port += 1

    return lock_resource('port', get_port, lock_dir, services_log, lock_resource_timeout)


def get_free_display(lock_dir, services_log, lock_resource_timeout):
    """Get free display to listen on."""
    def get_display(bound_resources):
        display = 100
        while True:
            if bound_resources:
                display = max(bound_resources) + 1
            if os.path.exists('/tmp/.X{0}-lock'.format(display)):
                display += 1
                continue
            return display

    return lock_resource('display', get_display, lock_dir, services_log, lock_resource_timeout)


@pytest.fixture(scope='session')
def port_getter(request, lock_dir, services_log, lock_resource_timeout):
    """Lock getter function."""
    def get_port():
        """Lock a free port and unlock it on finalizer."""
        port = get_free_port(lock_dir, services_log, lock_resource_timeout)

        def finalize():
            unlock_port(port, lock_dir, services_log)
        request.addfinalizer(finalize)
        return port
    return get_port


@pytest.fixture(scope='session')
def display_getter(request, lock_dir, services_log, lock_resource_timeout):
    """Display getter function."""
    def get_display():
        """Lock a free display and unlock it on finalizer."""
        display = get_free_display(lock_dir, services_log, lock_resource_timeout)
        request.addfinalizer(lambda: unlock_display(display, lock_dir, services_log))
        return display
    return get_display



"""Django setttings helpers."""
import imp
import os
import sys

from importlib import import_module
from django.core.urlresolvers import clear_url_caches, set_urlconf
from django.template import context, base, loader
from django.utils import translation
from django.utils.translation import trans_real


def setup_django_settings(test_settings):
    """Override the enviroment variable and call the _setup method of the settings object to reload them."""
    os.environ['DJANGO_SETTINGS_MODULE'] = test_settings

    from django.conf import settings as django_settings

    # (re)setup django settings
    django_settings._setup()

    # reload settings
    reload_settings(django_settings)


def clean_django_settings():
    """Clean current django settings."""
    from django.conf import settings as django_settings

    del os.environ['DJANGO_SETTINGS_MODULE']
    django_settings._wrapped = None


def reload_settings(settings, databases=None):
    """Special routine to reload django settings.

    Including:
    urlconf module, context processor, templatetags settings, database settings.
    """
    if databases:
        settings.DATABASES.update(databases)

    # check if there's settings to reload
    if hasattr(settings, 'ROOT_URLCONF'):
        if settings.ROOT_URLCONF in sys.modules:
            imp.reload(sys.modules[settings.ROOT_URLCONF])
        import django
        if hasattr(django, 'setup'):
            django.setup()
        import_module(settings.ROOT_URLCONF)
        set_urlconf(settings.ROOT_URLCONF)
        settings.LANGUAGE_CODE = 'en'  # all tests should be run with English by default

        # Make the ConnectionHandler use the new settings, otherwise the ConnectionHandler will have old configuraton.
        from django.db.utils import ConnectionHandler
        import django.db
        from django.db.utils import load_backend
        import django.db.transaction
        import django.db.models
        import django.db.models.sql.query
        import django.core.management.commands.syncdb
        import django.db.models.sql.compiler
        import django.db.backends
        import django.db.backends.mysql.base
        import django.core.management.commands.loaddata

        # all modules which imported django.db.connections should be changed to get new ConnectionHanlder
        django.db.models.sql.compiler.connections = django.db.models.connections = \
            django.core.management.commands.loaddata.connections = \
            django.db.backends.connections = django.db.backends.mysql.base.connections = \
            django.core.management.commands.syncdb.connections = django.db.transaction.connections = \
            django.db.connections = django.db.models.base.connections = django.db.models.sql.query.connections = \
            ConnectionHandler(settings.DATABASES)

        # default django connection and backend should be also changed
        django.db.connection = django.db.connections[django.db.DEFAULT_DB_ALIAS]
        django.db.backend = load_backend(django.db.connection.settings_dict['ENGINE'])

        import django.core.cache
        django.core.cache.cache = django.core.cache._create_cache(django.core.cache.DEFAULT_CACHE_ALIAS)

        # clear django urls cache
        clear_url_caches()
        # clear django contextprocessors cache
        context._standard_context_processors = None
        # clear django templatetags cache
        base.templatetags_modules = None

        # reload translation files
        imp.reload(translation)
        imp.reload(trans_real)

        # clear django template loaders cache
        loader.template_source_loaders = None
        from django.template.loaders import app_directories
        imp.reload(app_directories)
    from django.template.base import get_templatetags_modules
    get_templatetags_modules.cache_clear()
    import django.apps
    import django
    import django.template
    django.template.engines.__dict__.pop('templates', None)
    django.template.engines._templates = None
    django.template.engines._engines = {}
    if django.apps.apps.ready:
        django.apps.apps.set_installed_apps(settings.INSTALLED_APPS)
    django.setup()
